;; Generated package description from wisi.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "wisi" "4.3.2.0.20240313.1" "Utilities for implementing an indentation/navigation engine using a generalized LR parser" '((emacs "25.3") (seq "2.20")) :commit "f2945aa7f63d9708281d90343a825b472d68d5e3" :authors '(("Stephen Leake" . "stephen_leake@stephe-leake.org")) :maintainer '("Stephen Leake" . "stephen_leake@stephe-leake.org") :keywords '("parser" "indentation" "navigation") :url "https://stephe-leake.org/ada/wisitoken.html")
