;; Set up building with gprbuild in elpa -*- no-byte-compile: t; lexical-binding: t; -*-
;;
;; Copyright (C) 2021-2024  Free Software Foundation, Inc.
;;
;; This file is part of GNU Emacs.
;;
;; GNU Emacs is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs.  If not, see <https://www.gnu.org/licenses/>.

(wisi-prj-select-cache
 "wisi.gpr"
 (create-ada-prj
  :name "wisi main"
  :compile-env
  (list
   (concat "SAL="       (expand-file-name "../../org.stephe_leake.sal"))
   (concat "WISITOKEN=" (expand-file-name "../../org.wisitoken"))
   ))
 "Makefile"
 )
